/*
 ============================================================================

  This file is part of the bbLean source code
  Copyright  2001-2003 The Blackbox for Windows Development Team
  Copyright  2004 grischka

  http://bb4win.sourceforge.net/bblean
  http://sourceforge.net/projects/bb4win

 ============================================================================

  bbLean and bb4win are free software, released under the GNU General
  Public License (GPL version 2 or later), with an extension that allows
  linking of proprietary modules under a controlled interface. This means
  that plugins etc. are allowed to be released under any license the author
  wishes. For details see:

  http://www.fsf.org/licenses/gpl.html
  http://www.fsf.org/licenses/gpl-faq.html#LinkingOverControlledInterface

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

 ============================================================================
*/ // generic Menu Items

#include "../BB.h"
#include "../Settings.h"
#include "MenuMaker.h"
#include "Menu.h"
#include <shellapi.h>

int MenuItem::center_justify = DT_CENTER;
int MenuItem::left_justify   = DT_LEFT;

int g_menu_item_count;

//const int icon_size=16;
//===========================================================================

//===========================================================================

MenuItem::MenuItem(const char* pszTitle)
{
	next        = NULL;

	m_pMenu     =
	m_pSubMenu  = NULL;

	m_nWidth    =
	m_nHeight   = 0;

	m_isNOP     =
	m_bActive   =
	m_isChecked = false;

	m_nSortPriority = M_SORT_NORMAL;
	m_ItemID    = MENUITEM_ID_NORMAL;

	m_pszTitle  = new_str(pszTitle);
	m_pszCommand = NULL;
	m_pszIcon = NULL;
	m_pidl      = NULL;

	++g_menu_item_count;
}

MenuItem::~MenuItem()
{
	UnlinkSubmenu();
	if (m_pszTitle) m_free(m_pszTitle);
	if (m_pszCommand) m_free(m_pszCommand);
	if (m_pszIcon) m_free(m_pszIcon);
	if (m_pidl) m_free(m_pidl);
	--g_menu_item_count;
}

void MenuItem::LinkSubmenu(Menu *pSubMenu)
{
	UnlinkSubmenu();
	m_pSubMenu = pSubMenu;
}

void MenuItem::UnlinkSubmenu(void)
{
	if (m_pSubMenu)
	{
		Menu *pSubMenu = m_pSubMenu;
		m_pSubMenu = NULL;
		if (this==pSubMenu->m_pParentItem)
			pSubMenu->LinkToParentItem(NULL);
		pSubMenu->decref();
	}
}

//===========================================================================
// Execute the command that is associated with the menu item
void MenuItem::Invoke(int button)
{
	// default implementation is a nop
}

//====================
void MenuItem::Key(UINT nMsg, WPARAM wParam)
{
	// default implementation is a nop
}

//====================
void MenuItem::ItemTimer(UINT nTimer)
{
	if (MENU_POPUP_TIMER == nTimer)
	{
		ShowSubMenu();
	}
}

//====================

// Mouse command MOUSEMOVE, LB_PRESSED, ...
void MenuItem::Mouse(HWND hwnd, UINT uMsg, DWORD wParam, DWORD lParam)
{
	int i = 0;
	switch(uMsg)
	{
		case WM_MBUTTONUP:
			i = INVOKE_MID;
			break;

		case WM_RBUTTONUP:
			i = INVOKE_RIGHT;
			break;

		case WM_LBUTTONUP:
			i = INVOKE_LEFT;
			if (m_pMenu->m_dblClicked) i |= INVOKE_DBL;
			break;

		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
			SetCapture(hwnd);

		case WM_MOUSEMOVE:
			Active(1);
			break;
	}

	if (i && m_bActive)// && 0 == m_isNOP)
	{
		Menu::g_DiscardMouseMoves = 5;
		Invoke(i);
	}
}

//====================

void MenuItem::Active(int active)
{
	bool new_active = active > 0;
	if (m_bActive == new_active) return;
	m_bActive = new_active;

	RECT r; GetItemRect(&r);
	InvalidateRect(m_pMenu->m_hwnd, &r, false);

	if (false == new_active)
	{
		m_pMenu->m_pActiveItem = NULL;
		return;
	}

	m_pMenu->UnHilite();
	m_pMenu->m_pActiveItem = this;
	if (active == 1) m_pMenu->set_timer(true);
}

//===========================================================================

void MenuItem::ShowSubMenu()
{
	if (m_pSubMenu == m_pMenu->m_pChild)
		return;

	m_pMenu->HideChild();

	if (NULL == m_pSubMenu || m_pSubMenu->IsPinned())
		return;

	m_pSubMenu->m_bIconized = false;
	m_pSubMenu->m_bOnTop = m_pMenu->m_bOnTop;
	m_pSubMenu->LinkToParentItem(this);
	m_pSubMenu->Validate();
	m_pSubMenu->Show(0, 0, 0);
	m_pSubMenu->HideChild(); // submenu might be on the screen already
	PostMessage(BBhwnd, BB_SUBMENU, 0, 0); //for bbsoundfx
}

//===========================================================================

//===========================================================================

void MenuItem::GetItemRect(RECT* r)
{
	r->top    = m_nTop;
	r->bottom = m_nTop + m_nHeight;
	r->left   = m_nLeft;
	r->right  = m_nLeft + m_nWidth;
}

void MenuItem::GetTextRect(RECT* r)
{
	r->top    = m_nTop;
	r->bottom = m_nTop  + m_nHeight;

	r->left   = m_nLeft + MenuInfo.nItemIndent + icon_margin;
	r->right  = m_nLeft + m_nWidth - MenuInfo.nItemIndent + icon_margin;
// 	if(m_pszCommand && !strncmp(m_pszCommand,"@BBCore.exec ",13) && Settings_menuicon){
// 		r->left += icon_margin;
// 		r->right += icon_margin;
// 	}
}

const char* MenuItem::GetDisplayString(void)
{
	if (Settings_menusBroamMode
		&& (MENUITEM_ID_CI & m_ItemID)
		&& m_pszCommand
		&& m_pszCommand[0] == '@')
		return m_pszCommand;

	return m_pszTitle;
}

//===========================================================================

void MenuItem::Measure(HDC hDC, SIZE *size)
{
	const char *title = GetDisplayString();
	if (m_isNOP && m_ItemID == MENUITEM_ID_NORMAL && 0 == title[0])
	{
		// empty [nop]: height = 66%
		size->cx = 0;
		size->cy = MenuInfo.nItemHeight * 5 / 8;
		return;
	}

	RECT r = { 0, 0, 0, 0 };
	DrawText(hDC, title, -1, &r, DT_MENU_MEASURE_STANDARD);
	size->cx = r.right;
	size->cy = MenuInfo.nItemHeight;
//	if(m_pszCommand && !strncmp(m_pszCommand,"@BBCore.exec ",13)){
//		size->cy = icon_size;
//		size->cx += icon_size+2;
//	}
}

//===========================================================================

void MenuItem::Paint(HDC hDC)
{
	RECT rect;

	COLORREF cr0 = mStyle.MenuFrame.TextColor;
	COLORREF sc = mStyle.MenuFrame.ShadowColor;
	bool lit = false;

	if (m_bActive && 0 == (m_isNOP & (MI_NOP_TEXT | MI_NOP_SEP)) && (0 == (m_isNOP & MI_NOP_DISABLED) || m_pSubMenu))
	{
		// draw hilite bar
		GetItemRect(&rect);
		StyleItem *pSI = &mStyle.MenuHilite;
		MakeStyleGradient(hDC, &rect, pSI, pSI->bordered);
		cr0 = pSI->TextColor;
		sc = pSI->ShadowColor;
		lit = true;
	}
	else
	if (m_isNOP & MI_NOP_DISABLED)
	{
		cr0 = mStyle.MenuFrame.disabledColor;
	}

	// draw menu item text
	const char *title = GetDisplayString();
	GetTextRect(&rect);
	if (0 == (m_ItemID & (~MENUITEM_ID_CI & (MENUITEM_ID_CIInt|MENUITEM_ID_CIStr))) || Settings_menusBroamMode){
		// draw text-shadow
		if (sc != CLR_INVALID){
			RECT ts;
			ts.left = rect.left + 1;
			ts.top = rect.top + 1;
			ts.bottom = rect.bottom + 1;
			ts.right = rect.right + 1;
			SetTextColor(hDC, sc);
			DrawText(hDC, title, -1, &ts, mStyle.MenuFrame.Justify | DT_MENU_STANDARD);
		}
		// draw text
		SetTextColor(hDC, cr0);
		DrawText(hDC, title, -1, &rect, mStyle.MenuFrame.Justify | DT_MENU_STANDARD);
	}else
	if (m_ItemID != MENUITEM_ID_CIStr){
		if (sc != CLR_INVALID){
			RECT ts;
			ts.left = rect.left + 1;
			ts.top = rect.top + 1;
			ts.bottom = rect.bottom + 1;
			ts.right = rect.right + 1;
			SetTextColor(hDC, sc);
			DrawText(hDC, title, -1, &ts, DT_CENTER | DT_MENU_STANDARD);
		}
		// draw text
		SetTextColor(hDC, cr0);
		DrawText(hDC, title, -1, &rect, DT_CENTER | DT_MENU_STANDARD);
	}
	// set back previous textColor
	if ((COLORREF)-1 != cr0) SetTextColor(hDC, cr0);

	if (m_isChecked) // draw check-mark
	{
		StyleItem *pSI = &mStyle.MenuHilite;
		bool pr = pSI->parentRelative;
		if (lit != pr) pSI = &mStyle.MenuFrame;

		int x, y = m_nTop + m_nHeight / 2;
		if ((FolderItem::m_nBulletPosition == DT_RIGHT) == (0 == (MENUITEM_ID_FOLDER & m_ItemID)))
			x = m_nLeft + m_nWidth - MenuInfo.nItemIndent / 2 - 1;
		else
			x = m_nLeft + MenuInfo.nItemIndent / 2;

		const int r = 3;
		rect.left   = x - r;
		rect.right  = x + r;
		rect.top    = y - r;
		rect.bottom = y + r;

		if (pr) MakeGradient(hDC, rect, B_SOLID, pSI->TextColor, 0, false, BEVEL_FLAT, 0, 0, 0, 0);
		else MakeStyleGradient(hDC, &rect, pSI, false);
	}

	if(m_pszIcon && *m_pszIcon){
		char *path = strdup(m_pszIcon); // menu icon path(ico, exe, dll, icl ...)
		int index = 0; //icon index

		if (strchr(path, ',')){ // if icon-index exists
			// separate path and icon with ","
			path = strtok(path, ",");
			index = atoi(strtok(NULL, ","));
		}
		
		// delete the double quotation included in the path 
		if (strchr(path, '\"')) path = strtok(path, "\"");

		HICON hIconSmall;
		ExtractIconEx(path, index, NULL, &hIconSmall, 1);
		if(hIconSmall){
			int icon_size = imin(MenuInfo.nItemIndent - 1, 16);
			DrawIconEx(hDC,rect.left - MenuInfo.nItemIndent - icon_margin + 1, rect.top, hIconSmall, icon_size, icon_size, 0, 0, DI_NORMAL);
			DestroyIcon(hIconSmall);
		}
		free(path);
	}
}

//===========================================================================
